import dgl
import torch as th
import torch.nn.functional as F
from model import GAT, GraphSage, GCN, JKNet
import time
from torchmetrics.functional import accuracy
import numpy as np
from sampler import sampler as sampler_
import sys
import argparse
import dgl.data as data
from args import get_args,get_args2,get_args3,get_args4

def eval(model, graph):
    with th.no_grad():
        model.eval()
        labels = graph.ndata['label']
        logits = model(graph)
        return accuracy(logits, labels).item()


parser = argparse.ArgumentParser('main')
parser.add_argument('--sample-type', type=str, default='uniform_sample',
                    help='the type of sample')
parser.add_argument('--p1',default=0.0, type=float, help='the value of parameter p1')
parser.add_argument('--p2',default=1.0, type=float, help='the value of parameter p2')
parser.add_argument('--q',default=0.5, type=float, help='the value of parameter q')
parser.add_argument('--lambda_',default=1.0, type=float, help='the value of parameter lambda')
parser.add_argument('--uniform-ratio', default='1.0', type=float, help='the ratio of uniform sampling')
parser.add_argument('--model-name', type=str,
                    choices=['gcn', 'gat', 'graphsage', 'jknet'],
                    help='the name of model')
parser.add_argument("--gpu", type=int, default=-1,
                    help="which GPU to use. Set -1 to use CPU.")
parser.add_argument('--dataset', type=str, default='cora_full',
                    choices=['cora_full', 'amazon', 'coauthor'],
                    help='the name of dataset')
parser.add_argument("--out-head", type=int, default=1,
                    help="number of output attention heads")
parser.add_argument("--heads", type=str, default='8',
                    help="list of the numbers of heads splited by '-' ")
parser.add_argument("--hidden-dims", type=str, default='8',
                    help="list of the numbers of heads splited by '-' ")
parser.add_argument("--residual", action="store_true", default=False,
                    help="use residual connection")
parser.add_argument("--bias", action="store_true", default=False,
                    help="use bias")
parser.add_argument("--feat-drop", type=float, default=.0,
                    help="input feature dropout")
parser.add_argument("--attn-drop", type=float, default=.0,
                    help="attention dropout")
parser.add_argument('--negative-slope', type=float, default=0.2,
                    help="the negative slope of leaky relu")
parser.add_argument("--epochs", type=int, default=100,
                    help="number of training epochs")
parser.add_argument("--lr", type=float, default=0.005,
                    help="learning rate")
parser.add_argument('--weight-decay', type=float, default=5e-4,
                    help="weight decay")
args_model = parser.parse_args()
if args_model.model_name == 'gat':
    args = get_args()
elif args_model.model_name == 'graphsage':
    args = get_args2()
elif args_model.model_name == 'gcn':
    args = get_args3()
elif args_model.model_name == 'jknet':
    args = get_args4()
if args.gpu > 0:
    device = 'cuda:0'
else:
    device = 'cpu'
print('dataset: {}'.format(args.dataset))
if args.dataset == 'cora_full':
    dataset = data.CoraFullDataset()
    graph = dataset[0]
    num_nodes = graph.num_nodes()
    n_classes = int(70)

elif args.dataset == 'amazon':
    dataset = data.AmazonCoBuyComputerDataset()
    graph = dataset[0]
    num_nodes = graph.num_nodes()
    n_classes = dataset.num_classes

elif args.dataset == 'coauthor':
    dataset = data.CoauthorCSDataset()
    graph = dataset[0]
    num_nodes = graph.num_nodes()
    n_classes = dataset.num_classes

index = np.arange(num_nodes)
index = np.random.permutation(index)
train_val_bound = 0.7
val_test_bound = 0.85
train_id = index[:int(num_nodes * train_val_bound)]
val_id = index[int(num_nodes * train_val_bound):int(num_nodes * val_test_bound)]
test_id = index[int(num_nodes * val_test_bound):]

g_train = graph.subgraph(train_id)
g_val = graph.subgraph(val_id)
g_test = graph.subgraph(test_id)

del graph

g_val = dgl.remove_self_loop(g_val)
g_val = dgl.add_self_loop(g_val)

g_test = dgl.remove_self_loop(g_test)
g_test = dgl.add_self_loop(g_test)

g_val = g_val.to(device)
g_test = g_val.to(device)

num_feats = g_train.ndata['feat'].shape[1]

if args_model.model_name == 'gat':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    heads = args.heads.split('-')
    heads = [int(head) for head in heads]

    model = GAT(num_feats,
                hidden_dims,
                n_classes,
                heads,
                args.out_head,
                F.elu,
                args.feat_drop,
                args.attn_drop,
                args.negative_slope,
                args.residual,
                args.bias)
elif args_model.model_name == 'graphsage':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    model = GraphSage(num_feats,
                      hidden_dims,
                      n_classes,
                      dropout=args.feat_drop)
elif args_model.model_name == 'gcn':
    hidden_dims = args.hidden_dims.split('-')
    hidden_dims = [int(hidden_dim) for hidden_dim in hidden_dims]
    model = GCN(num_feats,
                hidden_dims,
                n_classes,
                dropout=args.feat_drop)
elif args_model.model_name == 'jknet':
    model = JKNet(in_dim=num_feats,
                  hid_dim=args.hid_dim,
                  out_dim=n_classes,
                  num_layers=args.num_layers,
                  mode=args.mode,
                  dropout=args.dropout).to(device)
print(model)
model.to(device)

loss_fcn = th.nn.CrossEntropyLoss()
if args_model.model_name in ['gat', 'gcn', 'graphsage']:
    optimizer = th.optim.Adam(
        model.parameters(), lr=args.lr, weight_decay=args.weight_decay)
elif args_model.model_name == 'jknet':
    optimizer = th.optim.Adam(model.parameters(), lr=args.lr, weight_decay=args.lamb)

time_list = []

sample_type = args.sample_type
# uniform_sample    gds
sampler = sampler_(g_train)

uniform_sample_ratio = args.uniform_ratio
if_print = False
p1 = args.p1
p2 = args.p2
q = args.q
lambda_ = args.lambda_

print('model:{},sample_type:{}, uniform_sample_ratio:{}, para_p:{}, para_q:{}, quantile:{}'.format(
    args_model.model_name, sample_type, uniform_sample_ratio, p1, p2, q))

val_acc_list = []
acc_list = []
if sample_type == 'uniform_sample':
    subgraph, ratio = sampler.uniform_sample(uniform_sample_ratio)

elif sample_type == 'gds':
    subgraph, ratio = sampler.gds(p1=p1, p2=p2, q=q, lambda_=lambda_)

subgraph = dgl.add_self_loop(subgraph)
labels = subgraph.ndata['label'].to(device)
if args_model.model_name in ['gat', 'gcn', 'graphsage']:
    model.reset_parameters()
else:
    model.reset_params()
for epoch in range(args.epochs):
    start = time.time()

    model.train()
    subgraph = subgraph.to(device)
    logits = model(subgraph)
    loss = loss_fcn(logits, labels)
    optimizer.zero_grad()
    loss.backward()
    optimizer.step()

    train_acc = accuracy(logits, labels)
    time_list.append(time.time() - start)

    val_acc = eval(model, g_val)
    val_acc_list.append(val_acc)

    print("Epoch {:05d} | Time(s) {:.4f} | Loss {:.4f} | TrainAcc {:.4f} |"
          " ValAcc {:.4f} | ratio {:.4f}".
          format(epoch, np.mean(time_list), loss.item(), train_acc,
                 val_acc, ratio))

    acc = eval(model, g_test)
    if if_print:
        print('test accuracy:{}'.format(acc))
    acc_list.append(acc)
val_acc_list = np.array(val_acc_list)
result_acc = acc_list[val_acc_list.argmax()]
result_acc_list.append(result_acc)
print('times:{}, test acc:{} , ratio: {}'.format(i + 1, result_acc, ratio))
