import torch as th
from torch import nn
import torch.nn.functional as F
from dgl import function as fn
from dgl.nn.functional import edge_softmax
from dgl.base import DGLError
from dgl.utils import expand_as_pair, check_eq_shape, dgl_warning
from dgl.nn.pytorch.conv import GraphConv


class Identity(nn.Module):
    def __init__(self):
        super(Identity, self).__init__()

    def forward(self, x):
        """Return input"""
        return x


class GATConv(nn.Module):
    def __init__(self,
                 in_feats,
                 out_feats,
                 num_heads,
                 feat_drop=0.,
                 attn_drop=0.,
                 negative_slope=0.2,
                 residual=False,
                 activation=None,
                 allow_zero_in_degree=False,
                 bias=True):
        super(GATConv, self).__init__()
        self._num_heads = num_heads
        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats
        self._allow_zero_in_degree = allow_zero_in_degree
        if isinstance(in_feats, tuple):  # 如果给的是一个元组，是要在二部图上进行学习，两个部分的特征维度可以不一样
            self.fc_src = nn.Linear(
                self._in_src_feats, out_feats * num_heads, bias=False)
            self.fc_dst = nn.Linear(
                self._in_dst_feats, out_feats * num_heads, bias=False)
        else:
            self.fc = nn.Linear(  # 将输入的特征维数变成输出维数x头的数量
                self._in_src_feats, out_feats * num_heads, bias=False)
        # attention向量拆分成两个部分
        self.attn_l = nn.Parameter(th.FloatTensor(size=(1, num_heads, out_feats)))  # th.FloatTensor可直接创建Float类型的tensor
        self.attn_r = nn.Parameter(th.FloatTensor(size=(1, num_heads, out_feats)))
        self.feat_drop = nn.Dropout(feat_drop)  # 对 feature dropout
        self.attn_drop = nn.Dropout(attn_drop)  # 对 attention dropout
        self.leaky_relu = nn.LeakyReLU(negative_slope)
        self.if_bias = bias
        if self.if_bias:
            self.bias = nn.Parameter(th.FloatTensor(size=(num_heads * out_feats,)))
        else:
            self.register_buffer('bias', None)  # 不把bias看成一个参数来训练
        if residual:  # 残差结构
            if self._in_dst_feats != out_feats:
                self.res_fc = nn.Linear(
                    self._in_dst_feats, num_heads * out_feats, bias=False)
            else:
                self.res_fc = Identity()
        else:
            self.register_buffer('res_fc', None)
        self.reset_parameters()  # 初始化参数
        self.activation = activation

    def reset_parameters(self):
        gain = nn.init.calculate_gain('relu')  # Return the recommended gain value for the given nonlinearity function.
        if hasattr(self, 'fc'):
            nn.init.xavier_normal_(self.fc.weight, gain=gain)
        else:
            nn.init.xavier_normal_(self.fc_src.weight, gain=gain)
            nn.init.xavier_normal_(self.fc_dst.weight, gain=gain)
        nn.init.xavier_normal_(self.attn_l, gain=gain)
        nn.init.xavier_normal_(self.attn_r, gain=gain)
        if self.if_bias:  # 有所修改
            nn.init.constant_(self.bias, 0)
        if isinstance(self.res_fc, nn.Linear):
            nn.init.xavier_normal_(self.res_fc.weight, gain=gain)

    def set_allow_zero_in_degree(self, set_value):
        r"""
        set_value : bool
            The value to be set to the flag.
        """
        self._allow_zero_in_degree = set_value

    def forward(self, graph, feat, get_attention=False):
        with graph.local_scope():
            if not self._allow_zero_in_degree:
                if (graph.in_degrees() == 0).any():
                    raise DGLError('There are 0-in-degree nodes in the graph, '
                                   'output for those nodes will be invalid. '
                                   'This is harmful for some applications, '
                                   'causing silent performance regression. '
                                   'Adding self-loop on the input graph by '
                                   'calling `g = dgl.add_self_loop(g)` will resolve '
                                   'the issue. Setting ``allow_zero_in_degree`` '
                                   'to be `True` when constructing this module will '
                                   'suppress the check and let the code run.')

            if isinstance(feat, tuple):
                h_src = self.feat_drop(feat[0])
                h_dst = self.feat_drop(feat[1])
                if not hasattr(self, 'fc_src'):
                    feat_src = self.fc(h_src).view(-1, self._num_heads, self._out_feats)
                    feat_dst = self.fc(h_dst).view(-1, self._num_heads, self._out_feats)
                else:
                    feat_src = self.fc_src(h_src).view(-1, self._num_heads, self._out_feats)
                    feat_dst = self.fc_dst(h_dst).view(-1, self._num_heads, self._out_feats)
            else:
                h_src = h_dst = self.feat_drop(feat)
                feat_src = feat_dst = self.fc(h_src).view(
                    -1, self._num_heads, self._out_feats)  # 输入的特征 -> 每一个头所要作用的特征
                if graph.is_block:
                    feat_dst = feat_src[:graph.number_of_dst_nodes()]
            # a^T [Wh_i || Wh_j] = a_l Wh_i + a_r Wh_j
            el = (feat_src * self.attn_l).sum(dim=-1).unsqueeze(-1)  # 每一个头根据相应的attention的权重加权组合
            er = (feat_dst * self.attn_r).sum(dim=-1).unsqueeze(-1)  # 每一个头根据相应的attention的权重加权组合
            graph.srcdata.update({'ft': feat_src, 'el': el})  # 源节点计算的结果，特征信息和左半部分attention结果
            graph.dstdata.update({'er': er})  # 目标节点需要的信息：右半部分attention结果
            # compute edge attention, el and er are a_l Wh_i and a_r Wh_j respectively.
            graph.apply_edges(fn.u_add_v('el', 'er', 'e'))  # 两部分attention结果加起来
            e = self.leaky_relu(graph.edata.pop('e'))  # 获取信息后不需要的信息可以直接扔掉
            # compute softmax
            graph.edata['a'] = self.attn_drop(edge_softmax(graph, e))
            # message passing
            graph.update_all(fn.u_mul_e('ft', 'a', 'm'),
                             fn.sum('m', 'ft'))
            rst = graph.dstdata['ft']
            # residual
            if self.res_fc is not None:
                resval = self.res_fc(h_dst).view(h_dst.shape[0], self._num_heads, self._out_feats)
                rst = rst + resval
            # bias
            if self.bias is not None:
                rst = rst + self.bias.view(1, self._num_heads, self._out_feats)
            # activation
            if self.activation:
                rst = self.activation(rst)

            if get_attention:
                return rst, graph.edata['a']
            else:
                return rst


class GAT(nn.Module):
    def __init__(self,
                 in_dim,
                 hidden_dims,
                 out_dim,
                 heads,
                 out_head,
                 activation,
                 feat_drop,
                 attn_drop,
                 negative_slope,
                 residual,
                 bias):

        super(GAT, self).__init__()
        assert len(hidden_dims) == len(heads)

        self.num_layers = len(hidden_dims) + 1
        self.layers = nn.ModuleList()
        self.activation = activation
        self.out_head = out_head

        if self.num_layers == 1:
            self.layers.append(GATConv(
                in_dim, out_dim, out_head,
                feat_drop, attn_drop, negative_slope, residual,
                False, self.activation, bias=bias))
        else:
            self.layers.append(GATConv(
                in_dim, hidden_dims[0], heads[0],
                feat_drop, attn_drop, negative_slope, residual,
                False, self.activation, bias=bias))

            for l in range(1, len(hidden_dims)):
                self.layers.append(GATConv(
                    hidden_dims[l - 1] * heads[l - 1], hidden_dims[l], heads[l],
                    feat_drop, attn_drop, negative_slope, residual, self.activation, bias=bias))

            self.layers.append(GATConv(
                hidden_dims[-1] * heads[-1], out_dim,
                self.out_head, feat_drop, attn_drop, negative_slope, residual, bias=bias))

    def reset_parameters(self):
        for layer in self.layers:
            layer.reset_parameters()

    def forward(self, g):
        h = g.ndata['feat'].float()
        for l in range(self.num_layers):
            h, self.attn = self.layers[l](g, h, get_attention=True)
            h = h.flatten(1)
        if self.out_head == 1:
            logits = h.squeeze(1)
        else:
            logits = h.mean(1)
        return logits


class GCNlayer(nn.Module):
    def __init__(self, in_dim, out_dim, activation=None, bias=False,
                 dropout=0., batch_norm=False):
        super(GCNlayer, self).__init__()
        if bias:
            self.bias = nn.Parameter(th.zeros(out_dim))
        else:
            self.bias = False
        self.weight = nn.Linear(in_dim, out_dim, bias=False)

        self.activation = activation
        self.dropout = nn.Dropout(dropout)
        self.batch_norm = batch_norm
        if batch_norm:
            self.offset = nn.Parameter(th.zeros(out_dim))
            self.scale = nn.Parameter(th.ones(out_dim))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_normal_(self.weight.weight)

    def forward(self, g, feats):
        with g.local_scope():
            g.ndata['h'] = self.dropout(feats)
            g.ndata['d'] = g.in_degrees()
            g.apply_edges(lambda edges: {'w': (edges.src['d']*edges.dst['d']).float().clamp(min=1).unsqueeze(1).pow(-0.5)})
            g.update_all(fn.u_mul_e('h', 'w', 'm'),  # 在每个节点处将邻居特征加和起来
                         fn.sum('m', 'h'))
            h = g.ndata.pop('h')
            if self.bias:
                h = self.weight(h) + self.bias
            else:
                h = self.weight(h)
            if self.batch_norm:
                mean = h.mean(dim=1).view(h.shape[0], 1)
                var = h.var(dim=1, unbiased=False).view(h.shape[0], 1) + 1e-9
                h = (h - mean) * self.scale * th.rsqrt(var) + self.offset
            return h


class GCN(nn.Module):
    def __init__(self, in_dim, hidden_dims, out_dim, activation=F.relu,
                 dropout=0., batch_norm=False, bias=False):
        super(GCN, self).__init__()
        self.layers = nn.ModuleList()
        self.num_layers = len(hidden_dims) + 1
        self.activation = activation

        if self.num_layers == 1:
            self.layers.append(GCNlayer(
                in_dim, out_dim, self.activation, bias, dropout, batch_norm))
        else:
            self.layers.append(GCNlayer(
                in_dim, hidden_dims[0], self.activation, bias, dropout, batch_norm))
            for l in range(1, len(hidden_dims)):
                self.layers.append(GCNlayer(
                    hidden_dims[l - 1], hidden_dims[l], self.activation, bias, dropout, batch_norm))
            self.layers.append(GCNlayer(
                hidden_dims[-1], out_dim, self.activation, bias, dropout, batch_norm))

    def reset_parameters(self):
        for layer in self.layers:
            layer.reset_parameters()

    def forward(self, g):
        h = g.ndata['feat'].float()
        for l in range(self.num_layers):
            h = self.layers[l](g, h)
            h = h.flatten(1)
        return h


class SAGEConv(nn.Module):

    def __init__(self,
                 in_feats,
                 out_feats,
                 aggregator_type,
                 feat_drop=0.,
                 bias=True,
                 norm=None,
                 activation=None):
        super(SAGEConv, self).__init__()

        self._in_src_feats, self._in_dst_feats = expand_as_pair(in_feats)
        self._out_feats = out_feats
        self._aggre_type = aggregator_type
        self.norm = norm
        self.feat_drop = nn.Dropout(feat_drop)
        self.activation = activation
        if aggregator_type == 'pool':
            self.fc_pool = nn.Linear(self._in_src_feats, self._in_src_feats)
        if aggregator_type == 'lstm':
            self.lstm = nn.LSTM(self._in_src_feats, self._in_src_feats, batch_first=True)
        if aggregator_type != 'gcn':
            self.fc_self = nn.Linear(self._in_dst_feats, out_feats, bias=False)
        self.fc_neigh = nn.Linear(self._in_src_feats, out_feats, bias=False)
        if bias:
            self.bias = nn.parameter.Parameter(th.zeros(self._out_feats))
        else:
            self.register_buffer('bias', None)
        self.reset_parameters()

    def reset_parameters(self):
        gain = nn.init.calculate_gain('relu')
        if self._aggre_type == 'pool':
            nn.init.xavier_uniform_(self.fc_pool.weight, gain=gain)
        if self._aggre_type == 'lstm':
            self.lstm.reset_parameters()
        if self._aggre_type != 'gcn':
            nn.init.xavier_uniform_(self.fc_self.weight, gain=gain)
        nn.init.xavier_uniform_(self.fc_neigh.weight, gain=gain)

    def _compatibility_check(self):
        if not hasattr(self, 'bias'):
            dgl_warning("You are loading a GraphSAGE model trained from a old version of DGL, "
                        "DGL automatically convert it to be compatible with latest version.")
            bias = self.fc_neigh.bias
            self.fc_neigh.bias = None
            if hasattr(self, 'fc_self'):
                if bias is not None:
                    bias = bias + self.fc_self.bias
                    self.fc_self.bias = None
            self.bias = bias

    def _lstm_reducer(self, nodes):
        m = nodes.mailbox['m']  # (B, L, D)
        batch_size = m.shape[0]
        h = (m.new_zeros((1, batch_size, self._in_src_feats)),
             m.new_zeros((1, batch_size, self._in_src_feats)))
        _, (rst, _) = self.lstm(m, h)
        return {'neigh': rst.squeeze(0)}

    def forward(self, graph, feat, edge_weight=None):
        self._compatibility_check()
        with graph.local_scope():
            if isinstance(feat, tuple):
                feat_src = self.feat_drop(feat[0])
                feat_dst = self.feat_drop(feat[1])
            else:
                feat_src = feat_dst = self.feat_drop(feat)
                if graph.is_block:
                    feat_dst = feat_src[:graph.number_of_dst_nodes()]
            msg_fn = fn.copy_src('h', 'm')
            if edge_weight is not None:
                assert edge_weight.shape[0] == graph.number_of_edges()
                graph.edata['_edge_weight'] = edge_weight
                msg_fn = fn.u_mul_e('h', '_edge_weight', 'm')

            h_self = feat_dst
            if graph.number_of_edges() == 0:
                graph.dstdata['neigh'] = th.zeros(
                    feat_dst.shape[0], self._in_src_feats).to(feat_dst)
            lin_before_mp = self._in_src_feats > self._out_feats

            # Message Passing
            if self._aggre_type == 'mean':
                graph.srcdata['h'] = self.fc_neigh(feat_src) if lin_before_mp else feat_src
                graph.update_all(msg_fn, fn.mean('m', 'neigh'))  # gcn与mean聚合相比，考虑了节点自己的特征
                h_neigh = graph.dstdata['neigh']
                if not lin_before_mp:
                    h_neigh = self.fc_neigh(h_neigh)
            elif self._aggre_type == 'gcn':
                check_eq_shape(feat)
                graph.srcdata['h'] = self.fc_neigh(feat_src) if lin_before_mp else feat_src
                if isinstance(feat, tuple):  # heterogeneous
                    graph.dstdata['h'] = self.fc_neigh(feat_dst) if lin_before_mp else feat_dst
                else:
                    graph.dstdata['h'] = graph.srcdata['h']
                graph.update_all(msg_fn, fn.sum('m', 'neigh'))
                # divide in_degrees
                degs = graph.in_degrees().to(feat_dst)
                h_neigh = (graph.dstdata['neigh'] + graph.dstdata['h']) / (
                        degs.unsqueeze(-1) + 1)
                if not lin_before_mp:
                    h_neigh = self.fc_neigh(h_neigh)
            elif self._aggre_type == 'pool':
                graph.srcdata['h'] = F.relu(self.fc_pool(feat_src))
                graph.update_all(msg_fn, fn.max('m', 'neigh'))
                h_neigh = self.fc_neigh(graph.dstdata['neigh'])
            elif self._aggre_type == 'lstm':
                graph.srcdata['h'] = feat_src
                graph.update_all(msg_fn, self._lstm_reducer)
                h_neigh = self.fc_neigh(graph.dstdata['neigh'])
            else:
                raise KeyError('Aggregator type {} not recognized.'.format(self._aggre_type))

            # GraphSAGE GCN does not require fc_self.
            if self._aggre_type == 'gcn':
                rst = h_neigh
            else:
                rst = self.fc_self(h_self) + h_neigh

            # bias term
            if self.bias is not None:
                rst = rst + self.bias

            # activation
            if self.activation is not None:
                rst = self.activation(rst)
            # normalization
            if self.norm is not None:
                rst = self.norm(rst)
            return rst


class GraphSage(nn.Module):
    def __init__(self, in_dim, hidden_dims, out_dim, agg_type='gcn', activation=F.relu, dropout=0., bias=False,
                 norm=None):
        super(GraphSage, self).__init__()
        self.layers = nn.ModuleList()
        self.num_layers = len(hidden_dims) + 1
        self.activation = activation

        if self.num_layers == 1:
            self.layers.append(SAGEConv(
                in_dim, out_dim, agg_type, dropout, bias, norm, activation))
        else:
            self.layers.append(SAGEConv(in_dim, hidden_dims[0], agg_type, dropout, bias, norm, activation))
            for l in range(1, len(hidden_dims)):
                self.layers.append(SAGEConv(
                    hidden_dims[l - 1], hidden_dims[l], agg_type, dropout, bias, norm, activation))
            self.layers.append(SAGEConv(
                hidden_dims[-1], out_dim, agg_type, dropout, bias, norm, activation))

    def reset_parameters(self):
        for layer in self.layers:
            layer.reset_parameters()

    def forward(self, g):
        h = g.ndata['feat'].float()
        for l in range(self.num_layers):
            h = self.layers[l](g, h)
            h = h.flatten(1)
        return h


class JKNet(nn.Module):
    def __init__(self,
                 in_dim,
                 hid_dim,
                 out_dim,
                 num_layers=1,
                 mode='cat',
                 dropout=0.):
        super(JKNet, self).__init__()

        self.mode = mode
        self.dropout = nn.Dropout(dropout)
        self.layers = nn.ModuleList()
        self.layers.append(GraphConv(in_dim, hid_dim, activation=F.relu))
        for _ in range(num_layers):
            self.layers.append(GraphConv(hid_dim, hid_dim, activation=F.relu))

        if self.mode == 'cat':
            hid_dim = hid_dim * (num_layers + 1)
        elif self.mode == 'lstm':
            self.lstm = nn.LSTM(hid_dim, (num_layers * hid_dim) // 2, bidirectional=True, batch_first=True)
            self.attn = nn.Linear(2 * ((num_layers * hid_dim) // 2), 1)

        self.output = nn.Linear(hid_dim, out_dim)
        self.reset_params()

    def reset_params(self):
        self.output.reset_parameters()
        for layers in self.layers:
            layers.reset_parameters()
        if self.mode == 'lstm':
            self.lstm.reset_parameters()
            self.attn.reset_parameters()

    def forward(self, g):
        feats = g.ndata['feat']
        feat_lst = []
        for layer in self.layers:
            feats = self.dropout(layer(g, feats))
            feat_lst.append(feats)

        if self.mode == 'cat':
            out = torch.cat(feat_lst, dim=-1)
        elif self.mode == 'max':
            out = torch.stack(feat_lst, dim=-1).max(dim=-1)[0]
        else:
            # lstm
            x = torch.stack(feat_lst, dim=1)
            alpha, _ = self.lstm(x)
            alpha = self.attn(alpha).squeeze(-1)
            alpha = torch.softmax(alpha, dim=-1).unsqueeze(-1)
            out = (x * alpha).sum(dim=1)

        g.ndata['h'] = out
        g.update_all(fn.copy_u('h', 'm'), fn.sum('m', 'h'))

        return self.output(g.ndata['h'])